close all; clear all; clc;
%----------------------------------------------
%Estrutura Geral do Simulador MIL-STD-188-110C:
%----------------------------------------------
%TRANSMISSOR: (bits) --> CODIFICADOR CONVOLUCIONAL --> INTERLEAVER --> MAPEADOR GRAY --> (SCRAMBLER + INSERCAO DE PREAMBULO/MINIPROBES) --> MODULADOR 8-PSK --> FORMATADOR PULSO --> (entrada do canal)
%
%CANAL HF: (saida do transmissor) --> MODELO WATTERSON --> RUIDO AWGN --> (entrada do receptor)
%
%RECEPTOR HARD: (saida do canal) --> (filtro recepcao)  --> (sincronizador quadro/simbolo) --> (equalizacao) --> (REMOCAO DE PREAMBULO/MINIPROBES + DESCRAMBLER) --> DEMODULADOR --> DEMAPEADOR GRAY --> DEINTERLEAVER --> (entrada do decodificador 'hard')
%
%RECEPTOR SOFT: (saida do canal) --> (filtro recepcao)  --> (sincronizador quadro/simbolo) --> (equalizacao) --> (REMOCAO DE PREAMBULO/MINIPROBES + DESCRAMBLER) --> (formacao de 'soft' bits) --> DEINTERLEAVER --> (entrada do decodificador 'soft')

%-----------------
%MIL-STD-188-110C:
%-----------------
%Variavel que ajusta o tamanho (duracao) do bloco de interleaving (entrelacamento).
%Opcoes sao de 0.6s (interleaving curto) e 4.8s (interleaving longo);
varMil110C.interLength = 4.8;

%Variavel que ajusta o numero de blocos de interleaving transmitidos;
varMil110C.nFrames     = 1;

%--------------------------
%Codificador convolucional:
%--------------------------
% O codificador convolucional utilizado por este simulador corresponde a um
%codigo proprietario. Entretanto, sua funcionalidade e equivalente a do
%objeto 'comm.convolutionalEncoder' do MATLAB. Este objeto pode ser chamado
%da sequinte maneira:
%
% h_conv_cod = comm.ConvolutionalEncoder(poly2trellis(7,[133 171]),'TerminationMethod','Truncated'),
%
%onde '7' e o constraint length (numero de registros de memoria do codificador-
%mae, incluindo o do bit de entrada no instante atual) e '[133 171]' representam
%os 2 polinomios geradores do codigo '1/2', em octal. Em seguida, os bits
%de entrada podem ser codificados fazendo
%
% CodOutInterIn = step(h_conv_cod, BitInCodIn).

%------------
%Equalizador:
%------------
%Variavel que ajusta o numero de coeficientes diretos do DFE. Deve ser um inteiro positivo;
varEqualizer.Nd = 16;

%Variavel que ajusta o numero de coeficientes de realimentacao do DFE. Deve ser um inteiro
%nao-negativo (para Nr=0, o equalizador passa a ser linear). Um ajuste adequado para este parametro
%tende a ser em torno do atraso entre raios (em simbolos) introduzido pelo canal;
varEqualizer.Nr = 5;

%Variavel que ajusta o 'tap' de referencia do equalizador. Assume valor minimo igual a 1 e valor
%maximo igual ao numero de coeficientes diretos do DFE (Nd). Este parametro determina a distribuicao
%dos 'taps' do filtro direto destinados ao processamento de precursores e poscursores. Um ajuste
%conservador (e subotimo) e o meio do filtro, ou seja, em torno de Nd/2. Ele se relaciona com o
%retardo introduzido pelo equalizador pela forma 'retardo_dfe = reftap-1';
varEqualizer.reftap = 11;

%Variavel que ajusta o fator de esquecimento do algoritmo Inverse-QR RLS. Assume valores entre zero
%e um. Representa um compromisso entre velocidade de convergencia (valores mais distantes de 1) e 
%estabilidade em regime (valores mais proximos de 1);
varEqualizer.lambda = 0.975;

%-------
%Filtro:
%-------
%Variavel que ajusta o fator de rolloff do filtro formatador de pulso no
%transmissor;
varShapping.rolloffFactor      = 0.35;

%Variavel que ajusta o atraso (em simbolos) do filtro formatador de pulso; 
varShapping.pulseDelay         = 5;

%Variavel que ajusta o fator de superamostragem implementado pelo filtro
%formatador de pulso. Retornara um erro se o valor adotado nao for tal que
%o produto 'AER x 2400 x oversamplingFactor' resulte um inteiro (sendo AER e
%o atraso (em segundos) entre os raios do canal Watterson);
varShapping.oversamplingFactor = 10;

%-----------------
%Watterson + AWGN:
%-----------------
%Variavel que seleciona o modelo de canal. Sao 5 opcoes, a primeira sendo o
%canal apenas AWGN: 'ideal'. As demais correspondem a 4 tipos de canal
%Watterson, cada um equivalendo a pares de valores de atraso entre raios (AER)
%e espalhamento Doppler (ED) distintos: canal 'bom' (AER = 0.0005s e ED =
%0.1Hz), canal 'moderado' (AER = 0.001s e ED = 0.5Hz), canal 'ruim' (AER =
%0.002s e ED = 1Hz) e canal 'muito_ruim' (AER = 0.005s e ED = 5Hz);
varWatterson.model  = 'ruim';

%OBS: varWatterson.EsNodB = Variavel que ajusta a razao sinal-ruido, em decibeis. 
%Corresponde a razao entre a energia media do simbolo PSK transmitido (Es) e a 
%densidade espectral de potencia do ruido (No). E ajustado dentro do loop da 
%simulacao de Monte Carlo pela variavel 'varSimulator.EsNodB';

%----------
%Simulador:
%----------
%Vetor com os valores de Es/No (em dB) assumidos pela variavel 'varWatterson.EsNodB'
%durante a simulacao de Monte Carlo:
varSimulator.EsNodB = 10:4:30;

%Vetor que seleciona as realizacoes desejadas para cada Es/No durante a simulacao 
%de Monte Carlo:
varSimulator.ensemble = 1:5000;

%mil110CParametros = carrega parametros gerais da forma de onda selecionada.
%Retorna um struct 'mil110C' que orienta o funcionamento dos demais componentes
%da cadeia de transmissao:
mil110C          = mil110CParametros(varMil110C.interLength);

MSE = zeros(2 * 1440 * varMil110C.interLength / 0.6, numel(varSimulator.EsNodB));
BERNoDecod = zeros(numel(varSimulator.EsNodB), 1);
BERDecod = zeros(numel(varSimulator.EsNodB), 1);
for ind_EsNodB = 1:numel(varSimulator.EsNodB)

	%Ajusta a razao sinal-ruido:
	varWatterson.EsNodB = varSimulator.EsNodB(ind_EsNodB);

	for ind_ensemble = 1:numel(varSimulator.ensemble)

		%Ajusta a semente da realizacao:
		rng(varSimulator.ensemble(ind_ensemble));

		%mil110CTransmissor = simula o transmissor MIL-STD-188-110C. Retorna uma struct
		%'TxOut' com os seguintes vetores:
		%
		%BitInCodIn = bits de dados transmitidos. Entrada do codificador
		%convolucional. Para quadros de interleaving curto (duração 0.6s), será um
		%vetor de multiplos de 1440 bits; para quadros de interleaving longo
		%(duração 4.8s), sera de multiplos de 11520 bits;
		%
		%CodOutInterIn = bits de saida do codificador convolucional e de entrada 
		%do interleaver. Como o codigo convolucional utiliza taxa 1/2, o tamanho
		%deste vetor sera o dobro do vetor 'BitInCodIn' (multiplos de 2880 para
		%interleaving curto, e multiplos de 11520 para interleaving longo);
		%
		%InterOutMapIn = bits de saida do interleaver e de entrada do mapeador 
		%Gray. Seu tamanho é o mesmo do vetor 'CodOutInterIn';
		%
		%MapOutMultiplexIn = simbolos PSK, inteiros, de saida do mapeador Gray e de
		%entrada do formador de quadro (multiplexador). O tamanho deste vetor
		%corresponde ao do vetor 'InterOutMapIn' dividido pelo numero de bits por
		%simbolo. Como a forma de onda 2400bps utiliza modulação 8-PSK, seu tamanho
		%será um terço da de 'InterOutMapIn';
		%
		%ComplexPSKSymbol = corresponde a versao modulada, em simbolos PSK
		%complexos, do 'MapOutMultiplexIn'. Seu tamanho será o mesmo da do
		%vetor 'MapOutMultiplexIn';
		%
		%ModConstellation = constelacao PSK utilizada para modular os simbolos de
		%entrada do modulador;
		%
		%ProbeInfo = vetor binário que indica a distribuicao/posicao de probes ao
		%longo do quadro transmitido ('1' indica que o simbolo naquela posicao e 
		%conhecido (probe), enquanto '0' indica que o simbolo naquela posicao e 
		%desconhecido no receptor). Para interleaving curto, as primeiras 1440
		%posicoes correspondem ao preambulo (para interleaving longo, este valor e de
		%11520). Apos estes valores, as posicoes seguintes correspondem a
		%intercalacoes de quadros de dados com dados de probes de treinamento. Para
		%a forma de onde 2400bps, o tamanho dos quadros de dados e de 32 simbolos,
		%enquanto o dos de probes e de 16 simbolos.
		%
		%MultiplexOut = saida do transmissor. Simbolos PSK, complexos, de saida 
		%do modulador, incluindo preambulo, probes e dados. Entrada do modulo 
		%formatador de pulso (segundo a norma, raiz de coseno levantado). Seu
		%tamanho e o mesmo do vetor ProbeInfo;
		TxOut            = mil110CTransmissor(mil110C, varMil110C.nFrames);

		%mil110CFiltroTransmissor = simula o siltro de saida do transmissor
		%(formatador de pulso). Retorna um struct 'ShappingOut' com as seguintes
		%saidas:
		%
		%filter = filtro raiz de coseno levantado utilizado na formatacao dos
		%pulsos. Seu tamanho se da pela relacao '2 x pulseDelay x oversamplingFactor + 1';
		%
		%out = dados de saida do formatador de pulso. Pulsos ocupando 
		%largura de banda (1 + rolloffFactor) x 1200 (banda base). Seu
		%tamanho corresponde ao tamanho do vetor 'MultiplexOut' multiplicado pelo
		%fator de superamostragem (oversamplingFactor), acrescido ainda das
		%amostras de retardo (pulseDelay x oversamplingFactor) introduzidas pelo 
		%filtro no começo e no fim do vetor;
		ShappingOut      = mil110CFiltroTransmissor(TxOut.MultiplexOut, varShapping);

		%mil110CCanalWatterson = simula o canal HF ionosferico, segundo o modelo de
		%canal proposto por (Watterson, 1970). Retorna um struct 'WattersonOut' com
		%as seguintes saidas:
		%
		%raios = matriz com 2 colunas, cada uma representando um dos raios 'Watterson'
		%ao longo da transmissao do sinal. Seu tamanho e o mesmo do vetor
		%'ShappingOut.out';
		%
		%AER = atraso entre os raios, em segundos;
		%
		%ED = espalhamento Doppler dos raios, em Hz;
		%
		%WattersonOutAwgnIn = sinal convoluido com os raios e, portanto, com presenca
		%de ISI. Seu tamanho corresponde ao do vetor 'ShappingOut.out', acrescido
		%do atraso (em amostras) entre os dois raios do canal Watterson (AER x 2400
		% x oversamplingFactor);
		%
		%noise = ruido AWGN adicionado ao sinal 'WattersonOutAwgnIn'. Seu tamanho e
		%o mesmo do vetor 'WattersonOutAwgnIn';
		%
		%AwgnOut = dados de saida do emulador de canal Watterson.
		%Corresponde ao sinal com presenca de ISI e contaminado com AWGN. Seu
		%tamanho e o mesmo do vetor 'WattersonOutAwgnIn';
		WattersonOut     = mil110CCanalWatterson(ShappingOut.out, varWatterson, varShapping);

		%mil110CFiltroCasado = simula o filtro de recepcao. Para este simulador
		%basico proposto, trata-se de um filtro casado ao filtro formatador de
		%pulso. Retorna um struct 'MatchedFilterOut' com as seguintes saidas:
		%
		%filter = filtro utilizado na filtragem do sinal de recepcao. Neste caso,
		%ele e semelhante ao 'ShappingOut.filter';
		%
		%out = dados de saida do filtro de recepcao. Neste caso, corresponde ao sinal
		%confinado novamente na largura de banda (1 + rolloffFactor) x 1200
		%(banda base). Seu tamanho corresponde ao tamanho do vetor 'AwgnOut'
		%acrescido das amostras de retardo introduzidas pelo filtro no começo e no
		%final do vetor;
		MatchedFilterOut = mil110CFiltroCasado(WattersonOut.AwgnOut, varShapping);

		%mil110CSincronismoQuadro = simula um sincronizador de quadro simples. O modelo
		%Watterson para o canal HF é um filtro FIR com 2 componentes principais. 
		%Para recuperar o sinal transmitido, o receptor precisa sincronizar a cadeia
		%de recepção em um desses dois componentes. Esse papel é realizado por alguma
		%técnica de sincronismo de quadro, no início da recepção do sinal. Para este
		%trabalho, considerou-se que o sincronismo de quadro encontra o componente de
		%maior módulo e sincroniza a cadeia de recepção em torno deste cursor. O outro
		%componente passa a ser um pós-cursor (caso o primeiro componente seja sincronizado)
		%ou um pré-cursor (caso o segundo o seja). Este sincronizador simples
		%retorna um numero, que e uma estimativa da amostra onde se identifica o
		%comeco da transmissao. Na ausencia de canal, ele retornara o valor
		%correspondente às amostras de retardo dos filtros de transmissao e de
		%recepcao, mais um;
		FrameStartSample = mil110CSincronismoQuadro(MatchedFilterOut.out, mil110C, varShapping);

		%mil110CSincronismoSimbolo = simula um sincronizador de simbolo (ponto de
		%amostragem) simples. Neste caso, ele apenas recebe o ponto de inicio da
		%transmissao e amostra o sinal de saida do filtro de recepcao em intervalos
		%de 'varShapping.oversamplingFactor' amostras. Retorna um struct 'SyncSymbolOut'
		%com a seguinte saida:
		%
		%out = dados de saida do sincronizador de simbolo. Corresponde a um vetor de
		%simbolos PSK, complexos. Na ausencia de canal, deve ser identico ao vetor
		%'TxOut.MultiplexOut'. Na presenca de canal, sera uma versao distorcida de 
		%'MultiplexOut', e o seu tamanho correspondera ao de 'MultiplexOut'
		%acrescido de alguns simbolos no final, correspondentes ao espalhamento de
		%atraso provocado pelo canal;
		SyncSymbolOut    = mil110CSincronismoSimbolo(MatchedFilterOut.out, FrameStartSample, varShapping);

		%mil110CEqualizador = simula o equalizador do receptor. Retorna um struct
		%'EqualizerOut' com a seguinte saida:
		%
		%out = vetor com simbolos 8-PSK complexos. Na ausencia de canal, deve ser 
		%similar ao vetor 'TxOut.MultiplexOut'. Na presenca de canal, sera uma 
		%versao detectada de 'MultiplexOut'. Seu tamanho e identico ao de 'MultiplexOut';
		EqualizerOut = mil110CEqualizador(SyncSymbolOut.out, TxOut.MultiplexOut, TxOut.ProbeInfo, TxOut.ModConstellation, varEqualizer);

		%mil110CDemultiplexador = simula o demultiplexador do receptor. Sua funcao
		%e separar os simbolos PSK de dados do preambulo e dos probes, além de fazer
		%o descrambling (desembaralhamento) dos simbolos. Retorna um struct
		%'DemultiplexOut' com a seguinte saida:
		%
		%out = vetor com simbolos PSK, complexos, de dados. Na ausencia de canal,
		%deve ser similar ao vetor 'TxOut.ComplexPSKSymbols'. Seu tamanho e
		%identico ao de 'ComplexPSKSymbols';
		DemultiplexOut   = mil110CDemultiplexador(EqualizerOut.out, mil110C);

		%mil110CDemodulador = simula o demodulador PSK e o demapeador Gray no
		%receptor. Retorna um struct 'DemodulatorOut' com as seguintes saidas:
		%
		%DemodOutDemapIn = simbolos PSK, inteiros, de saida do demodulador e de
		%entrada do demapeador Gray. Seu tamanho e identico ao do vetor 
		%'DemultiplexOut.out';
		%
		%DemapOut = 'hard' bits de saida do demapeador Gray, ainda sob efeito de
		%entrelacamento (interleaving) e codificacao convolucional. Na ausencia de
		%canal, deve ser identico ao vetor 'TxOut.InterOutMapIn'. Seu tamanho e
		%identico ao de 'InterOutMapIn';
		DemodulatorOut   = mil110CDemodulador(DemultiplexOut.out, mil110C);

		%mil110CDeinterleaver = simula o deinterleaver no receptor. Retorna um
		%struct com a seguinte saida:
		%
		%out = vetor com bits ('hard', quando a entrada for 'hard' bits; ou 'soft',
		%quando a entrada for 'soft' bits) desentrelacados, mas ainda 
		%codificados. Na ausencia de canal e quando consistir de 'hard' bits,
		%deve ser identico ao vetor 'TxOut.CodOutInterIn'. Seu tamanho e identico 
		%ao de 'CodOutInterIn';
		DeinterleaverOut = mil110CDeinterleaver(DemodulatorOut.DemapOut, mil110C);

		%mil110CViterbi = simula o decodificador convolucional no receptor. Retorna
		%um struct com a seguinte saida:
		%
		%out = vetor com bits decodificados. Na ausencia de canal, deve ser identico
		%ao vetor 'TxOut.BitInCodIn'. Seu tamanho e identico ao de 'BitInCodIn';
		ViterbiOut = mil110CViterbi(DeinterleaverOut.out, mil110C);

		%Acumula os resultados instantaneos:
		MSE(:,ind_EsNodB) = MSE(:,ind_EsNodB) + abs(EqualizerOut.out - TxOut.MultiplexOut).^2;
		BERNoDecod(ind_EsNodB,1) = BERNoDecod(ind_EsNodB,1) + sum(DeinterleaverOut.out ~= TxOut.CodOutInterIn) / numel(TxOut.CodOutInterIn);
		BERDecod(ind_EsNodB,1) = BERDecod(ind_EsNodB,1) + sum(ViterbiOut.out ~= TxOut.BitInCodIn) / numel(TxOut.BitInCodIn);

		%Mostra o percentual da simulacao:
		clc; disp([num2str((((ind_EsNodB-1)*numel(varSimulator.ensemble)+ind_ensemble)/...
		(numel(varSimulator.ensemble)*numel(varSimulator.EsNodB)))*100),'%']);

	end

	%Calcula os resultados da simulacao de Monte Carlo:
	MSE(:,ind_EsNodB) = MSE(:,ind_EsNodB)./numel(varSimulator.ensemble);
	BERNoDecod(ind_EsNodB,1) = BERNoDecod(ind_EsNodB,1)./numel(varSimulator.ensemble);
	BERDecod(ind_EsNodB,1) = BERDecod(ind_EsNodB,1)./numel(varSimulator.ensemble);

end

%-----------
%Resultados:
%-----------
%Erro quadratico medio na saida do equalizador:
figure;
plot(10*log10(MSE(:,end)));
xlabel('Simbolo'); 
ylabel('dB');
title('Erro quadratico medio na saida do DFE');
legend(['Es/No = ', num2str(varSimulator.EsNodB(end)), ' dB']);
grid on;

%Taxa de erros de bit na saida do equalizador (BERNoDecod) e do decodificador (BERDecod):
figure;
semilogy(varSimulator.EsNodB, BERNoDecod, '*-b');
hold on;
semilogy(varSimulator.EsNodB, BERDecod, 'o-r');
xlabel('Es/No (dB)'); 
ylabel('BER');
title('Taxas de erros de bit');
legend('Saida do DFE', 'Saida do Viterbi');
grid on;
